/*****************************************************************
 * Copyright (c) ParaSoft Corporation 1991.  All rights reserved.*
 * THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF ParaSoft       *
 * The copyright notice above does not evidence any              *
 * actual or intended publication of such source code.           *
 *****************************************************************/

static int print = 0;

#include "extern.h"

/**/U_DEFINE_STACK (key_list, int, 10)

#define LONGEST_KEYWORD	15

int key_current;

typedef struct Keyword {
	char *name;
	int type;
	struct Keyword *next;
} Keyword;

#define KEY_DOUBLE		1001
#define KEY_BLOCK		1002
#define KEY_ENDBLOCK	1004
#define KEY_GO			1005

static Keyword *hash_table [HASHSIZE];

static Keyword table [] = {
{"ALLOCATE", ALLOCATE, (Keyword*) 0},
{"ASSIGN", ASSIGN, (Keyword*) 0},
{"BACKSPACE", BACKSPACE, (Keyword*) 0},
/* ##BEG##, 9/19/91, SHU */
{"BLOCK", DIRBLOCK},
{"CYCLIC", DIRCYCLIC},
{"DECOMPOSITION", DIRDECOMPOSE},
{"DISTRIBUTE", DIRDISTRIBUTE},
/* ##END## */
/* ##BEG##, 10/27/91, ZEKI */
{"ALIGN", DIRALIGN},
{"WITH", DIRWITH},
/* ##END## */
{"CALL", CALL, (Keyword*) 0},
{"CLOSE", CLOSE, (Keyword*) 0},
{"CONTAINS", CONTAINS, (Keyword*) 0},
{"CONTINUE", CONTINUE, (Keyword*) 0},
{"CYCLE", CYCLE, (Keyword*) 0},
{"DEALLOCATE", DEALLOCATE, (Keyword*) 0},
{"ELSE", ELSE, (Keyword*) 0},
{"END", END, (Keyword*) 0},
{"ENTRY", ENTRY, (Keyword*) 0},
{"EQUIVALENCE", EQUIVALENCE, (Keyword*) 0},
{"EXIT", EXIT, (Keyword*) 0},
{"FORMAT", FORMAT, (Keyword*) 0},
{"INQUIRE", INQUIRE, (Keyword*) 0},
{"INTENT", INTENT, (Keyword*) 0},
{"NULLIFY", NULLIFY, (Keyword*) 0},
{"OPEN", OPEN, (Keyword*) 0},
{"PARAMETER", PARAMETER, (Keyword*) 0},
{"PRINT", PRINT, (Keyword*) 0},
{"PRIVATE", PRIVATE, (Keyword*) 0},
{"PROGRAM", KPROGRAM, (Keyword*) 0},
{"READ", READ, (Keyword*) 0},
{"RETURN", RETURN, (Keyword*) 0},
{"REWIND", REWIND, (Keyword*) 0},
{"SEQUENCE", SEQUENCE, (Keyword*) 0},
{"STOP", STOP, (Keyword*) 0},
{"WHERE", WHERE, (Keyword*) 0},
{"WRITE", WRITE, (Keyword*) 0},
{"FORALL", FORALL, (Keyword*) 0},
{"BLOCKDATA", BLOCK_DATA, (Keyword*) 0},
{"BLOCK", KEY_BLOCK, (Keyword*) 0},
{"DOUBLEPRECISION", DOUBLE_PRECISION, (Keyword*) 0},
{"DOUBLECOMPLEX", DOUBLE_COMPLEX, (Keyword*) 0},
{"DOUBLE", KEY_DOUBLE, (Keyword*) 0},
{"ELSEIF", ELSE_IF, (Keyword*) 0},
{"ELSEWHERE", ELSE_WHERE, (Keyword*) 0},
{"ENDBLOCKDATA", END_BLOCK_DATA, (Keyword*) 0},
{"ENDBLOCK", KEY_ENDBLOCK, (Keyword*) 0},
{"ENDDO", END_DO, (Keyword*) 0},
{"ENDFILE", END_FILE, (Keyword*) 0},
{"ENDFUNCTION", END_FUNCTION, (Keyword*) 0},
{"ENDIF", END_IF, (Keyword*) 0},
{"ENDINTERFACE", END_INTERFACE, (Keyword*) 0},
{"ENDMODULE", END_MODULE, (Keyword*) 0},
{"ENDPROGRAM", END_PROGRAM, (Keyword*) 0},
{"ENDSELECT", END_SELECT, (Keyword*) 0},
{"ENDSUBROUTINE", END_SUBROUTINE, (Keyword*) 0},
{"ENDTYPE", END_TYPE, (Keyword*) 0},
{"ENDWHERE", END_WHERE, (Keyword*) 0},
{"GOTO", GO_TO, (Keyword*) 0},
{"GO", KEY_GO, (Keyword*) 0},
{"INTERFACE", INTERFACE, (Keyword*) 0},
{"SELECT", SELECT, (Keyword*) 0},
{"USE", USE, (Keyword*) 0},
{"MODULE", MODULE, (Keyword*) 0},
{"IF", IF, (Keyword*) 0},
{"DO", DO, (Keyword*) 0},
{"INCLUDE", INCLUDE, (Keyword*) 0},
{"PAUSE", PAUSE, (Keyword*) 0},
{"CHARACTER", CHARACTER, (Keyword*) 0},
{"COMMON", COMMON, (Keyword*) 0},
{"COMPLEX", COMPLEX, (Keyword*) 0},
{"DATA", DATA, (Keyword*) 0},
{"DIMENSION", DIMENSION, (Keyword*) 0},
{"EXTERNAL", EXTERNAL, (Keyword*) 0},
{"IMPLICIT", IMPLICIT, (Keyword*) 0},
{"INTEGER", INTEGER, (Keyword*) 0},
{"INTRINSIC", INTRINSIC, (Keyword*) 0},
{"LOGICAL", LOGICAL, (Keyword*) 0},
{"LPIMP", LPIMP, (Keyword*) 0},
{"LPEQI", LPEQI, (Keyword*) 0},
{"NAMELIST", NAMELIST, (Keyword*) 0},
{"OPTIONAL", OPTIONAL, (Keyword*) 0},
{"POINTER", POINTER, (Keyword*) 0},
{"REAL", REAL, (Keyword*) 0},
{"SAVE", SAVE, (Keyword*) 0},
{"TARGET", KTARGET, (Keyword*) 0},
{"TYPE", TYPE, (Keyword*) 0},
{"RECURSIVE", RECURSIVE, (Keyword*) 0},
{"FUNCTION", FUNCTION, (Keyword*) 0},
{"SUBROUTINE", SUBROUTINE, (Keyword*) 0},
{"TO", TO, (Keyword*) 0},
{"CASE", CASE, (Keyword*) 0}
};

static int num_keywords;

static key_install (key)
Keyword *key;
{
	unsigned hashval;

	hashval = hash (key->name);
	key->next = hash_table [hashval];
	hash_table [hashval] = key;
}

key_init ()
{
	int i;

	num_keywords = sizeof (table)/sizeof (Keyword);

	for (i = 0; i < HASHSIZE; i++)
		hash_table [i] = (Keyword*) 0;

	for (i = 0; i < num_keywords; i++)
		key_install (table + i);
}

static Keyword *key_lookup (name)
char *name;
{
	Keyword *key;

	for (key = hash_table [hash (name)]; key != (Keyword*) 0; key = key->next)
		if (strcmp (name, key->name) == 0)
			return key;

	return (Keyword*) 0;
}

int is_not_xalnum (c)
char c;
{
	if ((c >= 'a' && c <= 'z')
	||  (c >= 'A' && c <= 'Z')
	||  (c >= '0' && c <= '9')
	||   c == '_')
		return 0;
	else
		return 1;
}

int is_not_xalpha (c)
char c;
{
	if ((c >= 'a' && c <= 'z')
	||  (c >= 'A' && c <= 'Z')
	||   c == '_')
		return 0;
	else
		return 1;
}

int is_xalpha (c)
char c;
{
	if ((c >= 'a' && c <= 'z')
	||  (c >= 'A' && c <= 'Z')
	||   c == '_')
		return 1;
	else
		return 0;
}

int iswhite (c)
char c;
{
	if (c == ' ' || c == '\t')  /*!!! add all possible whites */
		return 1;
	else
		return 0;
}

static char *skip_whites (line)
char *line;
{
	if (*line == '\n' || *line == '\0')
		return line;

	for (; *line != '\n' && *line != '\0'; line++)
		if (!iswhite (*line))
			return line;

	return line;
}

static char *next_keyword (line, size, key)
char *line;
int size;
char key [];
{
	int i;

	if (is_not_xalnum (line [size - 1]))
		return (char*) 0;

	for (i = 0; i < size; i++)
		if (islower (line [i]))
			key [i] = toupper (line [i]);
		else
			key [i] = line [i];

	key [size] = '\0';

	return key;
}

static char *skip_label_and_name (line)
char *line;
{
	char *aft_label;

	while (iswhite (*line))
		line++;

	if (isdigit (*line))
		while (isdigit (*++line))
			;

	while (iswhite (*line))
		line++;

	aft_label = line;

	while (*line != '\n' && *line != '\0') {
		switch (*line) {
		
		case '(':
		case ')':
		case ',':
			goto out;
			break;
			
		case ':':
			line++;

			if (source_form == FORM_FREE)
				line = skip_whites (line);

			if (*line == ':')  /* TypeSpec :: ... */
				goto out;
			else
				aft_label = line;
				goto out;

			break;
		}

		line++;
	}

out:

	if (source_form == FORM_FREE)
		aft_label = skip_whites (aft_label);

	return aft_label;
}

static int keyword (key_str)
char key_str [];
{
	int i;
	Keyword *key;

	if ((key = key_lookup (key_str)) != (Keyword*) 0) {
		U_stack_push_key_list (key->type);
		return key->type;
	}

	return 0;
}

static char *skip_if (line)
char *line;
{
	int pl = 0;

	for (;;)
		switch (*line++) {

		case '(':
			pl++;
			break;

		case ')':
			if (--pl == 0)
				return line;
			break;

		case '\n':
		case '\0':
			fatal ("unmatched parenthesis in IF statement");
			return (char*) 0;
			break;
		}
}

/* is ELSE if following it there are only alphanumeric characters and whites */

int is_not_else (line)
char *line;
{
	char c;

	for (; iswhite (*line); line++)
		;

	if (!U_strncmp_nocase ("WHERE", line, 5))
		return 1;

	for (; *line != '\n' && *line != '\0'; line++)
		if (!iswhite (*line) && is_not_xalnum (*line))
			return 1;
	
	return 0;
}

int is_not_do (line)
char *line;
{
	int pl = 0;

	if (*line != 'u' && *line != 'U')  /* deals with do_while for fixed format */
		return 0;

	for (;;)
		switch (*line++) {

		case '(':
			pl++;
			break;

		case ')':
			pl--;
			break;

		case ':':
			if (pl == 0)  /* TypeSpec, ... :: ... */
				return 1;
			break;

		case '=':
			return 0;

		case '\n':
		case '\0':
			return 1;
		}
}

int is_not_single (line)
char *line;
{
	line = skip_whites (line);

	if (*line == '\n' || *line == '')
		return 0;
	else
		return 1;
}

char *next_is_function (line)
char *line;
{
	char func [9];

	while (iswhite (*line))  /*!!! temporary fix for free form */
		line++;

	while (!is_xalpha (*line))  /* skip over '*10' in character*10 function name ... */
		line++;

	if (strlen (line) < 8)
		return (char*) 0;

	strncpy (func, line, 8);

	U_strupper (func);

	if (!strncmp (func, "FUNCTION", 8)
	&& (state [state_level] == 0 || se_interfaces_within_interfaces))
		return line;
	else
		return (char*) 0;
}

char *fix_assign (line)
char *line;
{
	char c;

	while (isdigit (*line++))
		;

	return --line;
}

static char *skip_digits (line)
char *line;
{
	if (*line == '\n' || *line == '\0')
		return line;

	for (; *line != '\n' && *line != '\0'; line++)
		if (!isdigit(*line))
			return line;

	return line - 1;
}

fix_character (line)
char *line;
{
	int state = 0;
	char *oline;

	line = skip_whites (line);

	if (*line != '*')
		return;

	oline = line;

	for (;;)
		switch (*line++) {
		case ':':
			return;
		case '\n':
		case '\0':
			goto fix;
		}

fix:
	line = oline + 1;
	line = skip_whites (line);

	if (*line != '(') {  /* CHARACTER*123 , */
		line = skip_digits (line);
		line = skip_whites (line);

		if (*line == ',')
			*line = '~';

		return;
	}

	for (;;line++)  /* CHARACTER*(123) , */
		switch (*line) {

		case '(':
			state++;
			break;

		case ')':
			state--;
			if (state == 0) {
				line = skip_whites (line + 1);

				if (*line == ',')
					*line = '~';

				return;
			}
			break;

		case '\n':
		case '\0':
			return;
		}
}

char *fix_implicit (line, size)
char *line;
int size;
{
	int pl = 0;
	char *mark;

	for (;;)
		switch (*line++) {

		case '(':
			pl++;

			mark = line;
			while (*--mark != '#')
				;

			for (mark = mark + 1; mark < line - 2; mark++)
				*mark = *(mark + 1);

			*(line - 2) = '#';
			*(line - 1) = ' ';
			keyword ("LPIMP");
			break;

		case ')':
			pl--;
			break;

		case ',':
			if (pl == 0)
				return line;
			break;

		case '\n':
		case '\0':
			return line;
		}
}

fix_implicit_do (line)
char *line;
{
	char c, *forw;
	int pl;

	if (line == CN)
		return;

	for (;;) {
		loop:

		switch (*line++) {

		case '(':
			forw = line;

			pl = 1;
			for (;;)
				switch (*forw++) {

				case '(':
					pl++;
					break;

				case ')':
					if (--pl == 0)
						goto loop;
					break;

				case '=':
					if (pl == 1) {
						*(line - 1) = '{';
						goto loop;
					}
					break;

				case '\n':
				case '\0':
					return;
				}
			break;

		case '\n':
		case '\0':
			return;
		}
	}
}

static int free_compact (line)
char *line;
{
	int len = 0;
	char *p = line;

	if (!iswhite (*p))
		return 0;

	while (iswhite (*p) && (*p != '\0' || *p != '\n')) {
		len++;
		p++;
	}

	if (!isalpha (*p))
		return 0;

	for (; *(line + len) != '\0'; line++)
		*line = *(line + len);

	*line = '\0';

	return 1;
}

key_find_keywords (line)  /*!!! needs to check for #'s in the line, and give error */
char *line;
{
	int size, i, key_type;
	char key [LONGEST_KEYWORD + 1], *tline;
	int in_implicit = 0;

	if (source_form == FORM_FIXED)
		key_eat_blanks (line);
	else
		key_eat_blanks_free (line);

find_stmt:

	if (key_is_assign (line)) {
		fix_implicit_do (line);
		return;
	}

	line = skip_label_and_name (line);

	size = 0;
	while (next_keyword (line, ++size, key)) {
		if ((key_type = keyword (key))) {
			switch (key_type) {


/* ##BEG##, 9/22/91, SHU */

			case DIRDISTRIBUTE:
				for (i = 1, line [0] = '#'; i < size; i++)
					line [i] = ' ';

				line = line + size;
				/* handle keywords BLOCK, CYCLIC, etc */
				fix_distribute(line);
				break;
/* ##END## */

/* ##BEG##, 10/27/91, ZEKI */

			case DIRALIGN:
				for (i = 1, line [0] = '#'; i < size; i++)
					line [i] = ' ';

				line = line + size;
				/* handle keywords WITH */
				fix_align(line);
				break;
/* ##END## */

			case IMPLICIT:
				in_implicit = 1;

				for (i = 1, line [0] = '#'; i < size; i++)
					line [i] = ' ';

				line = line + size;
				goto find_stmt;
				break;

			case ASSIGN:
				for (i = 1, line [0] = '#'; i < size; i++)
					line [i] = ' ';

				line = fix_assign (line + size);
				size = 0;
				goto continue_while;
				break;

			case KEY_DOUBLE:
			case KEY_BLOCK:
			case KEY_ENDBLOCK:
			case KEY_GO:
				if (source_form == FORM_FREE)
					free_compact (line + size);

				U_stack_pop_key_list ();
				goto continue_while;
				break;

			case CHARACTER:
				fix_character (line + size);
			case DOUBLE_PRECISION:
			case DOUBLE_COMPLEX:
			case INTEGER:
			case COMPLEX:
			case REAL:
			case LOGICAL:
				for (i = 1, line [0] = '#'; i < size; i++)
					line [i] = ' ';

				if (in_implicit)
					line = fix_implicit (line = line + size, size);
				else if ((tline = next_is_function (line + size)) != (char*) 0)
						line = tline;

				goto find_stmt;
				break;

			case IF:
			case FORALL:
				for (i = 1, line [0] = '#'; i < size; i++)
					line [i] = ' ';

				if ((line = skip_if (line)) != (char*) 0)
					goto find_stmt;

				break;

			case RECURSIVE:
				for (i = 1, line [0] = '#'; i < size; i++)
					line [i] = ' ';
				line = line + size;
				goto find_stmt;

				break;

			case DO:
				if (is_not_do (line + size)) {
					U_stack_pop_key_list ();
					goto continue_while;
				}
				else {
					for (i = 1, line [0] = '#'; i < size; i++)
						line [i] = ' ';
				}
				break;

			case ELSE:
				if (is_not_else (line + size)) {
					if (source_form == FORM_FREE)
						free_compact (line + size);
					
					U_stack_pop_key_list ();
					goto continue_while;
				}
				else {
					for (i = 1, line [0] = '#'; i < size; i++)
						line [i] = ' ';
				}
				break;

			case END:
				if (is_not_single (line + size)) {
					if (source_form == FORM_FREE)
						free_compact (line + size);

					U_stack_pop_key_list ();
					goto continue_while;
				}
				else {
					for (i = 1, line [0] = '#'; i < size; i++)
						line [i] = ' ';
				}

				break;

			default:
				for (i = 1, line [0] = '#'; i < size; i++)
					line [i] = ' ';
				break;
			}

			break;
		}

		continue_while:;
	}

	fix_implicit_do (line);
}

/* ##BEG##, 9/23/91, SHU */
fix_distribute(line)
char *line;
{
    char *t=line, tmp[10];
    int i;

    while (strlen(t) > 4 && *t != '\0')  {
    	switch (*t) {
	  case 'B':
	  case 'b':
		for (i = 0; i < 5; i++) tmp[i] = *(t+i);
		if (!strncmp (U_strupper (tmp), "BLOCK", 5)) {
		    for (i = 1, *t++='#'; i < 5; i++,t++) *t = ' ';
		    U_stack_push_key_list(DIRBLOCK);
		    }
	        else t++;
		break;
	  case 'C':
	  case 'c':
		for (i = 0; i < 6; i++) tmp[i] = *(t+i);
		if (!strncmp (U_strupper (tmp), "CYCLIC", 6)) {
		    for (i = 1, *t++='#'; i < 6; i++,t++) *t = ' ';
		    U_stack_push_key_list(DIRCYCLIC);
		    }
	        else t++;
		break;
	  default:
		t++;
		break;
	}
	}
}
/* ##END## */


/* ##BEG##, 10/27/91, ZEKI */
fix_align(line)
char *line;
{
    char *t=line, tmp[10];
    int i;

    while (strlen(t) > 3 && *t != '\0')  {
    	switch (*t) {
	  case 'W':
	  case 'w':
		for (i = 0; i < 4; i++) tmp[i] = *(t+i);
		if (!strncmp (U_strupper (tmp), "WITH", 4)) {
		    for (i = 1, *t++='#'; i < 4; i++,t++) *t = ' ';
		    U_stack_push_key_list(DIRWITH);
		    }
	        else t++;
		break;
	  default:
		t++;
		break;
	}
	}
}
/* ##END## */


key_eat_blanks_free (line)
char *line;
{
/*
	char *t = line, c;

	if (flag_new)
		return;

	while ((c = *line++) != '\0')
		if (!iswhite (c)) {
			*t++ = c;
			while ((c = *line++) != '\0')
				*t++ = c;
		}

	*t = '\0';
*/
}

key_eat_blanks (line)
char *line;
{
	char *t = line, c;
	int before_label = 1;

	if (flag_new)
		return;

	for (;;)
		switch (c = *line++) {

		case '\0':
			*t = '\0';
			return;

		default:
			if (!iswhite (c)) {
				if (before_label && !isdigit (c)) {
					before_label = 0;
/* ##BEG##, 9/19/91, SHU */
					/*  *t++ = ' '; */
/* ##END## */

					*t++ = ' ';
				}

				*t++ = c;
			}
			break;

		}
}

int key_is_assign (line)
char *line;
{
	int i, pl, state = 0;
	char c;

	for (i = 0, pl = 0; *line != '\0'; line++)
		switch (*line) {

		case '(':
			pl++;
			break;

		case ')':
			pl--;
			if (pl == 0)
				if (source_form == FORM_FIXED) {
					if (!is_not_xalnum (*(line + 1)))
						return 0;
				}
				else {
					line = skip_whites (line + 1);
					if (!is_not_xalnum (*line))
						return 0;
				}
			break;
		
		case '=':
			if (pl == 0)
				state = 1;
			break;

		case ':':
			if (pl == 0)
				return 0;
			break;

		case ',':
			if (pl == 0 && state == 1)
				return 0;
			break;
		}

	return state;
}

int key_list_top_fun ()
{
	return key_list_top;
}

int key_list_fun (i)
int i;
{
	return key_list [i];
}

char *key_check (node, name)
Node *node;
char *name;
{
	int len = strlen (name);
	char *name_upper = U_str_copy_upper (node->com.str);

	if (strncmp (name_upper, name, len))
		errsyn (node->com.str);

	if (len == strlen (node->com.str))
		return CN;
	else 
		return node->com.str + len;
}
